//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
//	Work2048	(ver 1.00)	(c) よよやん 2014
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
//コンパイルには、「dxlib311.lib」、「dx4d.d」が必要です。
//以下より、羽流布さんの配布している「dx4d」をダウンロードしてください。
//	http://dzware.com/notebook/dxlib/
import dx4d;
import std.stdio;
import std.string;
import std.conv;
import keyobs;
import rnd;
import keyconfig;
pragma(lib,"dxlib311.lib");

const ConsorFlg = false;		//コンソールに出力するか

void main() {
	dx_SetGraphMode(300,450,16);
	dx_ChangeWindowMode(true);
	dx_SetWindowText("work 2048");
	if (dx_DxLib_Init() == -1) throw new Error("DXライブラリの初期化に失敗");
	scope(exit){		//終了時処理
		static if (ConsorFlg) {
			writeln(game.ActionLog);
			writeln("score : ", game.score);
		}
		dx_DxLib_End();
	}
	
	new KeyTable;
	//auto KeyConfig = new _KeyConfig;
	auto game = new Work2048;
	//MainLoop
	while(true){
		dx_ClearDrawScreen;
		game.DrawAll;
		with(Keyobs){
			check;
			if (trigger(KEY_INPUT_ESCAPE)) return;	//別格
			KeyCheckPoint0:		//キーチェック
			for (int i; i < _KeyMap_max; i++){
				if (trigger(KeyConfig.keyid(i))) {		//i番目のKeyIdのキーが押されたかの判定
					//writeln(KeyConfig.keyid(i)," : ",KeyConfig.stroke(i));
					foreach(Stroke; KeyConfig.stroke(i)){		//実際の実行
						switch(Stroke){
							//基本操作
							//push
							case '7': game.push(0); break;
							case '4': game.push(1); break;
							case '1': game.push(2); break;
							//pop
							case '9': game.pop(0); break;
							case '6': game.pop(1); break;
							case '3': game.pop(2); break;
							//process
							case '5': game.process; break;
							//catch
							case '2': game.cth; break;
							//特殊コマンド
							case 'H':
								game.CommandHelp;
								break KeyCheckPoint0;
							case 'N':
								if (game.SeedSetting) game.ResetGame;
								break KeyCheckPoint0;
							case 'Q':
								return;
								break KeyCheckPoint0;
							case 'R':
								game.ResetGame;
								break KeyCheckPoint0;
							case 'T':
								game.Tutorial;
								break KeyCheckPoint0;
							default:
						}
					}
				}
			}
		}
		if(dx_ProcessMessage == -1) return;
		dx_WaitTimer(50);
	}
	return;
}
//ゲーム部分
class Work2048 {

	//キーテーブル
	private string KeyConfigMessage;

	private bool StartFlg;		//起動してからキーを押したか

	private bool ClearFlg;		//クリア済みフラグ
	enum _CongratulationsPos { x = 80, y = 140}

	enum _seed_max = 999999;
	enum _SeedPos { x = 200, y = 430 }

	private long score = 0;
	enum _ScorePos { x = 80, y = 430 }

	enum _stack_max = 16;		//スタックの最大サイズ
	enum _RegisterMax = 3; //レジスタの最大サイズ
	enum _streamlength = 10;	//ストリームの長さ

	//グラフィックイメージ
	private int[int] StackImage;
	private int StackCaseImage;
	private int[int] RegisterImage;
	private int RegisterCaseImage;
	private int[int] StreamImage;
	private int StreamCaseImage;

	immutable string[] TutorImageFile;		//チュートリアルイメージファイル
	int[] TutorImage;		//ハンドル
	//イメージ一つ当たりのサイズ
	enum _StackImageSize		 { x = 60,	 y = 20 }
	enum _RegisterImageSize	 { x = 120,	 y = 40 }
	enum _StreamImageSize		 { x = 20,	 y = 20 }

	//それぞれの描画位置
	//(スタックは最も下の要素の左上の座標)
	enum _StackPos 		{ x = 10,	 y = 420 }
	enum _RegisterPos { x = 160, y = 290 }
	enum _StreamPos 	{ x = 30,	 y = 10 }

	enum _InputAreaPos { x = 80, y = 260 }

	enum _rnd_max = 1000;		//乱数の最大値
	//各数字の出る確率
	enum _next_8 = 64;				
	enum _next_4 = _next_8 + 128;
	enum _next_2 = _next_4 + 256;

	//スタック、レジスタ、ストリームの変数
	private int[] stack;
	private int stackptr; 
	private int[_RegisterMax] register;
	private int[_streamlength] stream;
	private int streamptr;



	//雑用バッファ
	private int buf;
	private char[] strbuf;
	//描画カウンタ
	private uint DrawCnt;
	//ログ
	string ActionLog;
	//その他情報表示エリアの位置
	enum _InfoAreaPos { x = 80, y = 40}

	this(){
		Rnd.seed = Randomize;
		Rnd.seed = Rnd.get(_seed_max); ActionLog ~= text(Rnd.seed) ~ ";";
		Rnd.max = _rnd_max;
		//ストリーム初期化
		StreamInit();
		//イメージのロード
		StackImage		 = [1:0, 2:0, 4:0, 8:0, 16:0, 32:0, 64:0, 128:0, 256:0, 512:0, 1024:0, 2048:0];
		RegisterImage	 = [1:0, 2:0, 4:0, 8:0, 16:0, 32:0, 64:0, 128:0, 256:0, 512:0, 1024:0, 2048:0];
		StreamImage		 = [1:0, 2:0, 4:0, 8:0, 16:0, 32:0, 64:0, 128:0, 256:0, 512:0, 1024:0, 2048:0];
		StackImageLoad("data/StackImage.png","data/StackCase.png");
		RegisterImageLoad("data/RegisterImage.png","data/RegisterCase.png");
		StreamImageLoad("data/StreamImage.png","data/StreamCase.png");
		TutorImageFile = [
			"data/Tutorial/Tutor_Reg1.png",
			"data/Tutorial/Tutor_Reg2.png",
			"data/Tutorial/Tutor_Stk1.png",
			"data/Tutorial/Tutor_Stk2.png",
			"data/Tutorial/Tutor_Strm1.png",
			"data/Tutorial/Tutor_Strm2.png",
			"data/Tutorial/Tutor_Strm3.png",
			"data/Tutorial/Tutor_Score1.png",
			"data/Tutorial/Tutor_Score2.png",
			"data/Tutorial/Tutor_Seed1.png",
			"data/Tutorial/Tutor_Seed2.png",
		];
		LoadTutorialImage;
		//キー設定の読み込み
		buf = KeyConfig.LoadKeyMap("config/keymap.txt");
		if (buf < 0) {
			KeyConfigMessage = "Error in Line(" ~ text(-buf) ~ ")" ;//Error
		} else if (buf > 0) {
			KeyConfigMessage = text(buf) ~ " Keys Loaded" ;
		}
	}
	//ゲームの再起動
	void ResetGame(){
		StartFlg = false;
		ClearFlg = false;
		stackptr = 0; 
		register[] = 0;
		streamptr = 0;
		score = 0;
		Rnd.init;
		StreamInit();
		ActionLog.length = 0;		//ログ
		ActionLog ~= text(Rnd.seed) ~ ";";
	}

	//スタックイメージのロード
	void StackImageLoad(string ImageFile, string CaseImageFile){
		foreach(GraphHandle; StackImage){
			if (GraphHandle != 0) dx_DeleteGraph(GraphHandle);
		}
		int[12] GraphHandleBuf;
		if (dx_LoadDivGraph(ImageFile.ptr, 12, 1, 12, _StackImageSize.x, _StackImageSize.y, GraphHandleBuf.ptr) == -1 ) throw new Error("StackImageの読み込みに失敗");
		for(int i; i < 12; i++){
			StackImage[2^^i] = GraphHandleBuf[i];
		}
		if (StackCaseImage != 0) dx_DeleteGraph(StackCaseImage);
		StackCaseImage = dx_LoadGraph(CaseImageFile.ptr);
	}
	//レジスタイメージのロード
	void RegisterImageLoad(string ImageFile, string CaseImageFile){
		foreach(GraphHandle; RegisterImage){
			if (GraphHandle != 0) dx_DeleteGraph(GraphHandle);
		}
		int[12] GraphHandleBuf;
		if (dx_LoadDivGraph(ImageFile.ptr, 12, 1, 12, _RegisterImageSize.x, _RegisterImageSize.y, GraphHandleBuf.ptr) == -1 ) throw new Error("RegisterImageの読み込みに失敗");
		for(int i; i < 12; i++){
			RegisterImage[2^^i] = GraphHandleBuf[i];
		}
		if (RegisterCaseImage != 0) dx_DeleteGraph(RegisterCaseImage);
		RegisterCaseImage = dx_LoadGraph(CaseImageFile.ptr);
	}
	//ストリームイメージのロード
	void StreamImageLoad(string ImageFile, string CaseImageFile){
		foreach(GraphHandle; StreamImage){
			if (GraphHandle != 0) dx_DeleteGraph(GraphHandle);
		}
		int[4] GraphHandleBuf;
		if (dx_LoadDivGraph(ImageFile.ptr, 4, 1, 4, _StreamImageSize.x, _StreamImageSize.y, GraphHandleBuf.ptr) == -1 ) throw new Error("StreamImageの読み込みに失敗");
		for(int i; i < 4; i++){
			StreamImage[2^^i] = GraphHandleBuf[i];
		}
		if (StreamCaseImage != 0) dx_DeleteGraph(StreamCaseImage);
		StreamCaseImage = dx_LoadGraph(CaseImageFile.ptr);
	}
	//####################
	//描画
	//スタック
	void DrawStack(){
		dx_DrawGraph(_StackPos.x-5, _StackPos.y-_stack_max*(_StackImageSize.y+5)+_StackImageSize.y, StackCaseImage,true);
		for(int i; i < stackptr; i++){
			dx_DrawGraph(_StackPos.x, _StackPos.y-i*(_StackImageSize.y+5), StackImage[stack[i]], true);
		}
	}
	//レジスタ
	void DrawRegister(){
		dx_DrawGraph(_RegisterPos.x-5, _RegisterPos.y-5, RegisterCaseImage,true);
		for(int i; i < 3; i++){
			if (register[i] != 0) dx_DrawGraph(_RegisterPos.x, _RegisterPos.y + i*(_RegisterImageSize.y+5), RegisterImage[register[i]], true);
		}
	}
	//ストリーム
	void DrawStream(){
		dx_DrawGraph(_StreamPos.x-5, _StreamPos.y-5, StreamCaseImage,true);
		for(int i = streamptr; i < _streamlength; i++){
			dx_DrawGraph(_StreamPos.x+(i-streamptr)*(_StreamImageSize.x+5), _StreamPos.y, StreamImage[stream[i]], true);
		}
		for(int i; i < streamptr; i++){
			dx_DrawGraph(_StreamPos.x+(i+_streamlength-streamptr)*(_StreamImageSize.x+5), _StreamPos.y, StreamImage[stream[i]], true);
		}
	}
	//スコアの表示
	void DrawScore(){
		dx_DrawString(_ScorePos.x, _ScorePos.y, toStringz("Score:" ~ text(score)), 0xFFFFFF);
	}
	//シード値の表示
	void DrawSeed(){
		dx_DrawString(_SeedPos.x, _SeedPos.y, toStringz(("[" ~ text(Rnd.seed) ~ "]").dup), 0xFFFFFF);
	}
	//おめでとうの表示
	void DrawCongratulations(){
		dx_DrawString(_CongratulationsPos.x, _CongratulationsPos.y, toStringz("Congratulations!"), 0xFFFF00);
		dx_DrawString(_CongratulationsPos.x+20, _CongratulationsPos.y+20, toStringz("[N] NewGame (New Seed)"), 0xFFFFFF);
		dx_DrawString(_CongratulationsPos.x+20, _CongratulationsPos.y+40, toStringz("[R] Retry (Same Seed)"), 0xFFFFFF);
		dx_DrawString(_CongratulationsPos.x+20, _CongratulationsPos.y+60, toStringz("[Q],[ESC] Quit"), 0xBBBBBB);
	}
	//スタート時のメッセージ表示
	void DrawStartMessage(){
		dx_DrawString(_InfoAreaPos.x, _InfoAreaPos.y, toStringz("Welcome Work2048"), 0x00FF80);
		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+20, toStringz("[H] Help (Commands)"), 0xFFFFFF);
		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+40, toStringz("[T] Tutorial"), 0xFFFFFF);
		dx_DrawString(_InfoAreaPos.x, _InfoAreaPos.y+80, toStringz("KeyConfig:"), 0xAAFFAA);
		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+100, toStringz(KeyConfigMessage), 0xFFFFFF);
		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+220, toStringz("[Q],[ESC] Quit"), 0xBBBBBB);
		//dx_DrawString(_InfoAreaPos.x, _InfoAreaPos.y, toStringz(""), 0xFFFFFF);
	}
	//まとめ
	void DrawAll(){
		DrawStream;
		DrawRegister;
		DrawStack;
		DrawScore;
		DrawSeed;
		if (ClearFlg) DrawCongratulations;
		if (StartFlg == false) DrawStartMessage;
		DrawCnt++;
		if (DrawCnt == 2^^6*3^^5*5^^4*7^^3) DrawCnt = 0;
	}
	static if (ConsorFlg) {
	//####################
	//DrawProcs ForConsor
	void DrawStackConsor(){
		write(" [");
		for(int i; i < stackptr; i++){
			write(stack[i], " ");
		}
	}
	void DrawRegisterConsor(){
		write(" [", register[0], ",", register[1], ",", register[2], "] ");
	}
	void DrawStreamConsor(){
		write(" <<");
		for(int i = streamptr; i < _streamlength; i++){
			write(stream[i],",");
		}
		for(int i; i < streamptr; i++){
			write(stream[i],",");
		}
		write(">> ");
	}
	void DrawAllConsor(){
		DrawRegisterConsor;
		DrawStreamConsor;
		DrawStackConsor;
		writeln();
	}
	//####################
	}

	//プッシュ
	//レジスタからスタックに一つ積む
	void push(int index){
		if (ClearFlg) return;
		if (stackptr >= _stack_max) return;
		if (stack.length == stackptr) stack.length += 10;
		if (register[index] != 0) {
			stack[stackptr++] = register[index];
			register[index] = 0;
			ActionLog ~= text(7-3*index);		//in
			score -= 2; if (score < 1) score = 0;
			StartFlg = true;
		}
	}
	//ポップ
	//スタックからレジスタに一つ取り出す
	void pop(int index){
		if (ClearFlg) return;
		if (stackptr == 0) return;
		if (register[index] == 0) {
			register[index] = stack[--stackptr];
			ActionLog ~= text(9-3*index);		//out
			score -= 2; if (score < 1) score = 0;
			StartFlg = true;
		}
	}
	//プロセス
	//レジスタに同じ数字があれば、結合する
	void process(){
		if (ClearFlg) return;
		if (register[0] == register[1] && register[0] != 0) {
			register[0] *= 2;
			register[1] = 0;
			stream[streamptr] = nextnum();
			streamptr ++;
			if (streamptr == _streamlength) {
				streamptr = 0;
			}
			if (register[0] > 4) score += register[0] / 2;
		}else if (register[0] == register[2] && register[0] != 0) {
			register[0] *= 2;
			register[2] = 0;
			stream[streamptr] = nextnum();
			streamptr ++;
			if (streamptr == _streamlength) {
				streamptr = 0;
			}
			if (register[0] > 4) score += register[0] / 2;
		}else if (register[1] == register[2] && register[1] != 0) {
			register[1] *= 2;
			register[2] = 0;
			stream[streamptr] = nextnum();
			streamptr ++;
			if (streamptr == _streamlength) {
				streamptr = 0;
			}
			if (register[1] > 4) score += register[1] / 2;
		}else return;
		if ( (register[0]==2048) || (register[1]==2048) ) {		//2048達成
			score *= 2;
			static if (ConsorFlg) write("congratulations!!!");
			ClearFlg = true;
			///////////////
		}
		StartFlg = true;
		ActionLog ~= "5";		//make
	}
	//キャッチ
	//ストリームから一つスタックに積む
	void cth(){
		if (ClearFlg) return;
		if (stackptr >= _stack_max) return;
		if (register[0] + register[1] + register[2] ==0){
			if (stack.length == stackptr) stack.length += 10;
			stack[stackptr++] = stream[streamptr];
			stream[streamptr] = nextnum();
			streamptr++;
			if (streamptr == _streamlength) {
				streamptr = 0;
			}
		} else return;
		ActionLog ~= "2";		//catch
		StartFlg = true;
		score += 5;
	}
	//ストリームを初期化する。
	void StreamInit(){
		foreach(ref buf; stream){
			buf = nextnum();
		}
	}
	//次の数字を生成
	uint nextnum(){
		buf = Rnd.get;
		if (buf < _next_8) {
			return 8;
		}else if (buf < _next_4) {
			return 4;
		}else if (buf < _next_2) {
			return 2;
		}else{
			return 1;
		}
	}
	//シードの再設定	false:設定されなかった
	bool SeedSetting(){
		StartFlg = true;
		dx_ClearDrawScreen;
		DrawAll;
		dx_DrawString(_InputAreaPos.x, _InputAreaPos.y-40, toStringz("SeedSetting (0~" ~ text(_seed_max) ~ ")"), 0xFFFF88);
		dx_DrawString(_InputAreaPos.x+20, _InputAreaPos.y-20, toStringz("[ESC] Cancel"), 0xFFFFFF);
		buf = dx_KeyInputNumber(_InputAreaPos.x, _InputAreaPos.y, _seed_max, 0, true);
		Keyobs.check;
		if (buf == _seed_max + 1) {		//キャンセルしたとき
			return false;
		} else {
			if (buf == -1) throw new Error("シード値の取得に失敗"); else {
				Rnd.seed = buf;
			} 
			return true;
		}
	}
	//キー説明
	void CommandHelp(){
		StartFlg = true;
		dx_ClearDrawScreen;
		DrawAll;
		dx_DrawString(_InfoAreaPos.x, _InfoAreaPos.y, toStringz("Help (Commands)"), 0x80FF80);
		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+20, toStringz("[H] Help (this)"), 0xFFFFFF);
		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+40, toStringz("[N] NewGame (NewSeed)"), 0xFFFFFF);
		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+60, toStringz("[R] Reset (SameSeed)"), 0xFFFFFF);
		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+80, toStringz("[T] Tutorial"), 0xFFFFFF);

		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+160, toStringz("[Q],[ESC] Quit"), 0xBBBBBB);

		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+220, toStringz("Press [5]"), 0xFFF000);
		while(true){
			with(Keyobs){
				check;
				if (trigger(KEY_INPUT_SPACE)||trigger(KEY_INPUT_NUMPAD5)) break;
				if (trigger(KEY_INPUT_Q)||trigger(KEY_INPUT_ESCAPE)) return;
				if (trigger(KEY_INPUT_NUMPAD4)||trigger(KEY_INPUT_B)) break;
			}
			if(dx_ProcessMessage == -1) return;
			dx_WaitTimer(50);
		}
	}
	//チュートリアル
	void LoadTutorialImage(){
		//画像のロード
		TutorImage.length = TutorImageFile.length;
		for(int i; i < TutorImageFile.length; i++){
			if (TutorImage[i] == 0) {
				TutorImage[i] = dx_LoadGraph(TutorImageFile[i].ptr);
				if (TutorImage[i] == -1) throw new Error("チュートリアル画像の読み込みに失敗");
			}
		}
	}
	void Tutorial(){
		StartFlg = true;
		dx_ClearDrawScreen;
		DrawAll;
		//Next,Prev操作説明
		dx_DrawString(_InfoAreaPos.x, _InfoAreaPos.y, toStringz("Tutorial"), 0x0080FF);
		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+20, toStringz("[Space],[6] : Next"), 0xFFFFFF);
		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+40, toStringz("[B],[4] : Prev"), 0xFFFFFF);
		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+60, toStringz("[Q],[ESC] : Quit"), 0xBBBBBB);
		dx_DrawString(_InfoAreaPos.x+20, _InfoAreaPos.y+220, toStringz("Press [6] ->"), 0xFFFF00);
		while(true){
			with(Keyobs){
				check;
				if (trigger(KEY_INPUT_SPACE)||trigger(KEY_INPUT_NUMPAD6)) break;
				if (trigger(KEY_INPUT_Q)||trigger(KEY_INPUT_ESCAPE)) return;
			}
			if(dx_ProcessMessage == -1) return;
			dx_WaitTimer(50);
		}
		//描画
		for(int i; i < TutorImageFile.length; i++){
			dx_ClearDrawScreen;
			DrawAll;
			dx_DrawGraph(0,0,TutorImage[i],true);
			while(true){
				with(Keyobs){
					check;
					if (trigger(KEY_INPUT_SPACE)||trigger(KEY_INPUT_NUMPAD6)) break;
					if (trigger(KEY_INPUT_Q)||trigger(KEY_INPUT_ESCAPE)) return;
					if (trigger(KEY_INPUT_NUMPAD4)||trigger(KEY_INPUT_B)) {
						i -= 2; 
						if (i < -1) i = -1;
						break;
					}
				}
				if(dx_ProcessMessage == -1) return;
				dx_WaitTimer(50);
			}
		}
		return;
	}
}
